import pytest
from ruamel.yaml import YAML

from somesy.cff.writer import CFF


def test_cff_validate_accept(load_files, file_types):
    """Validate by loading the data cff file using the fixture."""
    load_files([file_types.CITATION])


def test_cff_validate_reject(tmp_path):
    """Test validating a CFF file."""
    cff_path = tmp_path / "CITATION.cff"

    # create a CFF file with a invalid version
    reject_cff_object = {
        "version": "1.4.0",
        "message": "If you use this software, please cite it using these metadata.",
        "type": "software",
    }
    yaml = YAML()
    yaml.dump(reject_cff_object, cff_path)

    # try to load the CFF file
    with pytest.raises(ValueError):
        CFF(cff_path)



from enum import Enum
from pathlib import Path
from typing import Any, Dict, Set, Tuple, Type

import pytest

from somesy.cff import CFF
from somesy.core.log import SomesyLogLevel, set_log_level
from somesy.core.models import Person, SomesyInput
from somesy.package_json.writer import PackageJSON
from somesy.pyproject import Pyproject


class FileTypes(Enum):
    POETRY = "poetry"
    SETUPTOOLS = "setuptools"
    CITATION = "citation"
    SOMESY = "somesy"
    PACKAGE_JSON = "package_json"


@pytest.fixture(scope="session", autouse=True)
def init_somesy_logger():
    set_log_level(SomesyLogLevel.DEBUG)


@pytest.fixture
def somesy_input() -> dict:
    """Return a somesy input instance."""
    return SomesyInput.from_input_file(Path("tests/data/somesy.toml"))


@pytest.fixture
def file_types() -> Type[FileTypes]:
    """Return a FileTypes instance."""
    return FileTypes


@pytest.fixture
def create_files(tmp_path):
    """Create file types with given file names with a dict input, return the folder location.

    Example:
    ```python
    files = [{FileTypes.POETRY: 'pyproject.poetry.toml'}, {FileTypes.SETUPTOOLS: 'pyproject.setuptools.toml'}, ...]
    # file_dir folder has the requested files
    file_dir = create_files(files)
    ```
    """

    def _create_files(files: Set[Tuple[FileTypes, str]]):
        for file_tuple in files:
            file_type, file_name = file_tuple
            if not isinstance(file_type, FileTypes):
                raise ValueError(f"Invalid file type: {file_type}")
            write_file_name = tmp_path / Path(file_name)

            read_file_path = Path("tests/data")
            read_file_name: Path = None
            # set file name as the input, if not given set to default
            if file_type == FileTypes.CITATION:
                read_file_name = read_file_path / Path("CITATION.cff")
            elif file_type == FileTypes.SETUPTOOLS:
                read_file_name = read_file_path / Path("pyproject.setuptools.toml")
            elif file_type == FileTypes.POETRY:
                read_file_name = read_file_path / Path("pyproject.toml")
            elif file_type == FileTypes.SOMESY:
                read_file_name = read_file_path / Path("somesy.toml")
            elif file_type == FileTypes.PACKAGE_JSON:
                read_file_name = read_file_path / Path("package.json")

            with open(read_file_name, "r") as f:
                content = f.read()
                with open(write_file_name, "w+") as f2:
                    f2.write(content)

        return tmp_path

    yield _create_files


@pytest.fixture
def load_files():
    """Read and load files to defined classes, return a dict with those class instances.

    Example:
    ```python
    files = [FileTypes.POETRY, FileTypes.SETUPTOOLS, ...]
    # file_instances dict has the instances of the requested files
    file_instances = load_files(files)
    ```
    """

    def _load_files(files: Set[FileTypes]):
        file_instances: Dict[FileTypes, Any] = {}
        for file_type in files:
            if not isinstance(file_type, FileTypes):
                raise ValueError(f"Invalid file type: {file_type}")

            read_file_name = Path("tests/data")
            if file_type == FileTypes.CITATION:
                read_file_name = read_file_name / Path("CITATION.cff")
                file_instances[file_type] = CFF(read_file_name)
            elif file_type == FileTypes.SETUPTOOLS:
                read_file_name = read_file_name / Path("pyproject.setuptools.toml")
                file_instances[file_type] = Pyproject(read_file_name)
            elif file_type == FileTypes.POETRY:
                read_file_name = read_file_name / Path("pyproject.toml")
                file_instances[file_type] = Pyproject(read_file_name)
            elif file_type == FileTypes.SOMESY:
                read_file_name = read_file_name / Path("somesy.toml")
                file_instances[file_type] = SomesyInput.from_input_file(read_file_name)
            elif file_type == FileTypes.PACKAGE_JSON:
                read_file_name = read_file_name / Path("package.json")
                file_instances[file_type] = PackageJSON(read_file_name)

        return file_instances

    yield _load_files


@pytest.fixture
def person() -> Person:
    p = {
        "given-names": "Jane",
        "email": "j.doe@example.com",
        "family-names": "Doe",
        "orcid": "https://orcid.org/0123-4567-8910",
    }
    ret = Person.model_validate(p)
    ret.set_key_order(list(p.keys()))  # custom order!
    return ret



import pytest

from somesy.cff.writer import CFF
from somesy.core.models import LicenseEnum, Person, ProjectMetadata


@pytest.fixture
def cff(load_files, file_types):
    files = load_files([file_types.CITATION])
    return files[file_types.CITATION]


def test_content_match(cff: CFF):
    assert cff.name == "test-package"
    assert cff.description == "This is a test package for demonstration purposes."
    assert cff.license == "MIT"
    assert len(cff.authors) == 1


def test_sync(cff: CFF, somesy_input: dict):
    cff.sync(somesy_input.project)
    assert cff.name == "testproject"
    assert cff.version == "1.0.0"


def test_save(tmp_path):
    # test save with default path
    file_path = tmp_path / "CITATION.cff"
    cff = CFF(file_path, create_if_not_exists=True)
    cff.save()
    assert file_path.is_file()

    # test save with custom path
    custom_path = tmp_path / "custom.cff"
    cff.save(custom_path)
    assert custom_path.is_file()


def test_from_to_person(person: Person):
    assert CFF._from_person(person) == person.model_dump(by_alias=True)

    p = CFF._to_person(CFF._from_person(person))
    assert p.full_name == person.full_name
    assert p.email == person.email
    assert p.orcid == str(person.orcid)


def test_person_merge(tmp_path, person: Person):
    def to_cff_keys(lst):
        return list(map(lambda s: s.replace("_", "-"), lst))

    cff_path = tmp_path / "CITATION.cff"
    cff = CFF(cff_path, create_if_not_exists=True)

    pm = ProjectMetadata(
        name="My awesome project",
        description="Project description",
        license=LicenseEnum.MIT,
        version="0.1.0",
        people=[person.model_copy(update=dict(author=True, publication_author=True))],
    )
    cff.sync(pm)
    cff.save()

    # check that serialization preserves key order
    # (load raw dict from yaml and see order of keys)
    dct = cff._yaml.load(open(cff_path))
    assert list(dct["authors"][0].keys()) == to_cff_keys(person._key_order)

    # jane becomes john -> modified person
    person1b = person.model_copy(
        update={"given_names": "John", "author": True, "publication_author": True}
    )

    # different Jane Doe with different orcid -> new person
    person2 = person.model_copy(
        update={
            "orcid": "https://orcid.org/4321-0987-3231",
            "email": "i.am.jane@doe.com",
            "author": True,
            "publication_author": True,
        }
    )
    # use different order, just for some difference
    person2.set_key_order(["given_names", "orcid", "family_names", "email"])

    # listed in "arbitrary" order in somesy metadata (new person comes first)
    pm.people = [person2, person1b]  # need to assign like that to keep _key_order
    print("people:", pm.people, sep="\n")
    cff.sync(pm)
    cff.save()

    # existing author order preserved
    assert cff.authors[0] == person1b.model_dump(
        by_alias=True, exclude={"author", "publication_author"}
    )
    assert cff.authors[1] == person2.model_dump(
        by_alias=True, exclude={"author", "publication_author"}
    )
    # existing author field order preserved
    dct = cff._yaml.load(open(cff_path, "r"))
    assert list(dct["authors"][0].keys()) == to_cff_keys(person1b._key_order)
    assert list(dct["authors"][1].keys()) == to_cff_keys(person2._key_order)

    # new person
    person3 = Person(
        **{
            "given_names": "Janice",
            "family_names": "Doethan",
            "email": "jane93@gmail.com",
            "author": True,
            "publication_author": True,
        }
    )
    # john has a new email address
    person1c = person1b.model_copy(update={"email": "john.of.us@qualityland.com"})
    # jane 2 is removed from authors, but added to maintainers
    person2.author = False
    person2.publication_author = False
    person2.maintainer = True
    # reflect in project metadata
    pm.people = [person3, person2, person1c]
    # sync to CFF file
    cff.sync(pm)
    cff.save()

    assert len(cff.authors) == 2
    assert len(cff.maintainers) == 1
    assert cff.authors[0] == person1c.model_dump(
        by_alias=True, exclude={"author", "publication_author"}
    )
    assert cff.authors[1] == person3.model_dump(
        by_alias=True, exclude={"author", "publication_author"}
    )
    dct = cff._yaml.load(open(cff_path, "r"))
    assert list(dct["authors"][0].keys()) == to_cff_keys(person1c._key_order)



"""package.json validation models."""
import re
from typing import List, Optional, Union

from pydantic import BaseModel, EmailStr, Field, field_validator
from typing_extensions import Annotated

from somesy.core.types import HttpUrlStr


class PackageAuthor(BaseModel):
    """Package author model."""

    name: Annotated[Optional[str], Field(description="Author name")]
    email: Annotated[Optional[EmailStr], Field(description="Author email")]
    url: Annotated[
        Optional[HttpUrlStr], Field(description="Author website or orcid page")
    ]


class PackageRepository(BaseModel):
    """Package repository model."""

    type: Annotated[Optional[str], Field(description="Repository type")]
    url: Annotated[str, Field(description="Repository url")]


class PackageLicense(BaseModel):
    """Package license model."""

    type: Annotated[Optional[str], Field(description="License type")]
    url: Annotated[str, Field(description="License url")]


NPM_PKG_AUTHOR = r"^(.*?)\s*(?:<([^>]+)>)?\s*(?:\(([^)]+)\))?$"
NPM_PKG_NAME = r"^(@[a-z0-9-~][a-z0-9-._~]*\/)?[a-z0-9-~][a-z0-9-._~]*$"
NPM_PKG_VERSION = r"^(?:0|[1-9]\d*)\.(?:0|[1-9]\d*)\.(?:0|[1-9]\d*)(?:-(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*)(?:\.(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*))*)?(?:\+([0-9a-zA-Z-]+(?:\.[0-9a-zA-Z-]+)*))?$"


class PackageJsonConfig(BaseModel):
    """Package.json config model."""

    model_config = dict(populate_by_name=True)

    name: Annotated[str, Field(description="Package name")]
    version: Annotated[str, Field(description="Package version")]
    description: Annotated[Optional[str], Field(description="Package description")]
    author: Annotated[
        Optional[Union[str, PackageAuthor]], Field(description="Package author")
    ]
    maintainers: Annotated[
        Optional[List[Union[str, PackageAuthor]]],
        Field(description="Package maintainers"),
    ] = None
    contributors: Annotated[
        Optional[List[Union[str, PackageAuthor]]],
        Field(description="Package contributors"),
    ] = None
    license: Annotated[
        Optional[Union[str, PackageLicense]], Field(description="Package license")
    ]
    repository: Annotated[
        Optional[Union[PackageRepository, str]], Field(description="Package repository")
    ]
    homepage: Annotated[
        Optional[HttpUrlStr], Field(description="Package homepage")
    ] = None
    keywords: Annotated[
        Optional[List[str]], Field(description="Keywords that describe the package")
    ] = None

    # convert package author to dict if it is a string
    @classmethod
    def convert_author(cls, author: str) -> PackageAuthor:
        """Convert author string to PackageAuthor model."""
        # parse author string to "name <email> (url)" format with regex
        author_match = re.match(NPM_PKG_AUTHOR, author)
        if not author_match:
            raise ValueError(f"Invalid author format: {author}")
        author_name = author_match[1]
        author_email = author_match[2]
        author_url = author_match[3]

        return PackageAuthor(name=author_name, email=author_email, url=author_url)

    @field_validator("name")
    @classmethod
    def validate_name(cls, v):
        """Validate package name."""
        if re.match(NPM_PKG_NAME, v) is None:
            raise ValueError("Invalid name")

        return v

    @field_validator("version")
    @classmethod
    def validate_version(cls, v):
        """Validate package version."""
        if re.match(NPM_PKG_VERSION, v) is None:
            raise ValueError("Invalid version")
        return v

    @field_validator("author")
    @classmethod
    def validate_author(cls, v):
        """Validate package author."""
        return cls.convert_author(v) if isinstance(v, str) else v

    @field_validator("maintainers", "contributors")
    @classmethod
    def validate_people(cls, v):
        """Validate package maintainers and contributors."""
        people = []
        for p in v:
            if isinstance(p, str):
                people.append(cls.convert_author(p))
            else:
                people.append(p)
        return people



import sys
from datetime import datetime
from pathlib import Path

import pytest

from somesy.core.core import discover_input
from somesy.core.models import ProjectMetadata, SomesyConfig
from somesy.core.types import ContributionTypeEnum, LicenseEnum


def test_discover_input(create_files, file_types, monkeypatch: pytest.MonkeyPatch):
    tmp_path = create_files(
        [
            (file_types.SOMESY, "somesy.toml"),
            (file_types.POETRY, "pyproject.toml"),
            (file_types.PACKAGE_JSON, "package.json"),
        ]
    )

    # Test 1: input is given and exists

    # 1.a somesy input file
    somesy_file = tmp_path / Path("somesy.toml")
    result = discover_input(somesy_file)
    assert result == somesy_file

    # 1.b pyproject input file
    pyproject_file = tmp_path / Path("pyproject.toml")
    result = discover_input(pyproject_file)
    assert result == pyproject_file

    # 1.c package.json file
    package_json_file = tmp_path / Path("package.json")
    result = discover_input(package_json_file)
    assert result == package_json_file

    # Test 2: input is given but does not exist, instead default exists and selected
    input_file = tmp_path / Path("somesy2.toml")
    result = discover_input(input_file)
    assert result == Path(".somesy.toml")

    # Test 3: should raise an error when given file or defaults dont exist
    if sys.platform == "win32":
        monkeypatch.chdir(Path("C:\\"))
    else:
        monkeypatch.chdir(Path("/tmp"))
    input_file = Path("/tmp/somesy.toml")
    with pytest.raises(FileNotFoundError):
        discover_input(input_file)


def test_somesy_input(somesy_input):
    # test config inputs
    assert isinstance(somesy_input.config, SomesyConfig)
    assert somesy_input.config.debug == True
    assert somesy_input.config.verbose == False
    assert somesy_input.config.show_info == False
    assert somesy_input.config.no_sync_cff == False
    assert somesy_input.config.cff_file == Path("CITATION.cff")
    assert somesy_input.config.no_sync_pyproject == False
    assert somesy_input.config.pyproject_file == Path("pyproject.toml")
    assert somesy_input.config.no_sync_codemeta == False
    assert somesy_input.config.codemeta_file == Path("codemeta.json")
    assert somesy_input.config.no_sync_package_json == False

    # test project inputs
    assert isinstance(somesy_input.project, ProjectMetadata)
    assert somesy_input.project.name == "testproject"
    assert somesy_input.project.version == "1.0.0"
    assert (
        somesy_input.project.description
        == "This is a test project for demonstration purposes."
    )
    assert somesy_input.project.keywords == ["test", "demo", "example"]
    assert somesy_input.project.license == LicenseEnum.MIT
    assert (
        str(somesy_input.project.repository) == "https://github.com/example/testproject"
    )
    assert str(somesy_input.project.homepage) == "https://example.com/testproject"
    assert len(somesy_input.project.people) == 3
    authors = somesy_input.project.authors()
    assert authors[0].family_names == "Doe"
    assert authors[0].given_names == "John"
    assert authors[0].email == "john.doe@example.com"
    assert str(authors[0].orcid) == "https://orcid.org/0000-0000-0000-0000"
    assert authors[0].contribution == "The main developer, maintainer, and tester."
    assert (
        authors[0].contribution_begin
        == datetime.strptime("2023-01-15", "%Y-%m-%d").date()
    )
    assert authors[0].contribution_types == [
        ContributionTypeEnum.maintenance,
        ContributionTypeEnum.code,
        ContributionTypeEnum.test,
        ContributionTypeEnum.review,
        ContributionTypeEnum.doc,
    ]

    # check publication author
    assert authors[1].publication_author == True



from json import dump

import pytest

from somesy.package_json import PackageJSON


def test_package_json_validate_accept(load_files, file_types):
    """Validate by loading the data package.json file using the fixture."""
    load_files([file_types.PACKAGE_JSON])


def test_package_json_validate(tmp_path):
    """Test validating a package.json file in both package_json and setuptools formats."""

    # create a package.json file in package.json format but with a invalid values
    reject_package_json_object = {
        "name": "somesy",
        "version": "abc",
        "authors": "John Doe",
    }
    invalid_package_json_path = tmp_path / "package.json"
    with open(invalid_package_json_path, "w+") as f:
        dump(reject_package_json_object, f)

    with pytest.raises(ValueError):
        PackageJSON(invalid_package_json_path)



"""Integration with codemetapy (to re-generate codemeta as part of somesy sync)."""
import contextlib
import logging
from pathlib import Path

from ..core.models import SomesyConfig
from .exec import gen_codemeta
from .utils import cff_codemeta_tempfile, update_codemeta_file

log = logging.getLogger("somesy")


def collect_cm_sources(conf: SomesyConfig):
    """Assemble list of inputs for codemetapy based on somesy config.

    Returns files that are supported by both somesy and codemetapy and are enabled for somesy.
    """
    cm_sources = []
    if (
        not conf.no_sync_pyproject
        and conf.pyproject_file is not None
        and conf.pyproject_file.is_file()
    ):
        cm_sources.append(conf.pyproject_file)
    # NOTE: we don't add CFF directly, because it must be handled separately
    # NOTE: add other suitable somesy targets / codemeta sources (except CFF and codemeta) here
    if (
        conf.no_sync_package_json
        and conf.package_json_file is not None
        and conf.package_json_file.is_file()
    ):
        cm_sources.append(conf.package_json_file)
    return cm_sources


def update_codemeta(conf: SomesyConfig):
    """Generate or update codemeta file based on sources that somesy supports.

    Returns True if file has been written, False if it was up to date.
    """
    cm_sources = collect_cm_sources(conf)

    # if cff file is given, convert it to codemeta tempfile and pass as extra input
    temp_cff_cm = contextlib.nullcontext(None)
    if not conf.no_sync_cff and conf.cff_file is not None:
        temp_cff_cm = cff_codemeta_tempfile(conf.cff_file)
        cm_sources.append(Path(temp_cff_cm.name))

    # run codemetapy
    with temp_cff_cm:
        cm_harvest = gen_codemeta(cm_sources)

    # NOTE: once codemetapy is fixed (still broken with 2.5.1), we should not need
    # update_codemeta_file + codemeta context dump + most of utils.py + their tests anymore
    # we'll first disable the workaround and see for a while if everything works
    # as expected, and if it does, the cleanup can be completed.
    #
    # save to file with same settings as used by codemetapy
    # with open(conf.codemeta_file, "w") as f:
    #     f.write(json.dumps(cm_harvest, indent=4, ensure_ascii=False, sort_keys=True))

    # check output and write file if needed (with workaround checking graph equivalence)
    return update_codemeta_file(conf.codemeta_file, cm_harvest)


__all__ = ["update_codemeta"]



import pytest
from tomlkit import dump

from somesy.pyproject import Pyproject


def test_poetry_validate_accept(load_files, file_types):
    """Validate by loading the data pyproject file using the fixture."""
    load_files([file_types.SETUPTOOLS])
    load_files([file_types.POETRY])


def test_poetry_validate(tmp_path):
    """Test validating a pyproject file in both poetry and setuptools formats."""

    # create a pyproject file in poetry format but with a invalid values
    reject_poetry_object = {
        "tool": {
            "poetry": {"name": "somesy", "version": "abc", "authors": ["John Doe <"]}
        }
    }
    invalid_poetry_path = tmp_path / "pyproject.toml"
    with open(invalid_poetry_path, "w+") as f:
        dump(reject_poetry_object, f)

    with pytest.raises(ValueError):
        Pyproject(invalid_poetry_path)

    # create a pyproject file in setuptools format but with a invalid values
    reject_setuptools_object = {
        "project": {"name": "somesy", "version": "abc", "authors": ["John Doe <"]}
    }
    with open(invalid_poetry_path, "w+") as f:
        dump(reject_setuptools_object, f)
    with pytest.raises(ValueError):
        Pyproject(invalid_poetry_path)



"""Utility functions for CLI commands."""
import logging
import traceback
from typing import Optional

import typer
import wrapt
from rich.pretty import pretty_repr

from somesy.core.core import discover_input
from somesy.core.log import SomesyLogLevel, get_log_level, set_log_level
from somesy.core.models import SomesyConfig, SomesyInput

logger = logging.getLogger("somesy")


@wrapt.decorator
def wrap_exceptions(wrapped, instance, args, kwargs):
    """Format and log exceptions for cli commands."""
    try:
        return wrapped(*args, **kwargs)

    except Exception as e:
        logger.error(f"[bold red]Error: {e}[/bold red]")
        logger.debug(f"[red]{traceback.format_exc()}[/red]")
        raise typer.Exit(code=1) from e


def resolved_somesy_input(**cli_args) -> SomesyInput:
    """Return a combined `SomesyInput` based on config file and passed CLI args.

    Will also adjust log levels accordingly.
    """
    # figure out what input file to use
    input_file = discover_input(cli_args.pop("input_file", None))

    # create config based on passed arguments
    passed_args = {k: v for k, v in cli_args.items() if v is not None}
    somesy_conf = SomesyConfig(input_file=input_file, **passed_args)

    # cli_log_level is None if the user did not pass a log level (-> "default")
    cli_log_level: Optional[SomesyLogLevel] = get_log_level()

    if cli_log_level is not None:
        # update log level flags if cli log level was set
        somesy_conf.update_log_level(cli_log_level)

    somesy_input: SomesyInput = somesy_conf.get_input()

    if cli_log_level is None:
        # no cli log level -> set it according to the loaded configuration
        set_log_level(somesy_input.config.log_level())

    logger.debug(
        f"Combined config (Defaults + File + CLI):\n{pretty_repr(somesy_input.config)}"
    )
    return somesy_input



from pathlib import Path

import pytest

from somesy.core.models import LicenseEnum, Person, ProjectMetadata
from somesy.pyproject.writer import Poetry, SetupTools


@pytest.fixture
def pyproject_poetry(load_files, file_types):
    files = load_files([file_types.POETRY])
    return files[file_types.POETRY]


@pytest.fixture
def pyproject_poetry_file(create_files, file_types):
    folder = create_files([(file_types.POETRY, "pyproject.toml")])
    return folder / Path("pyproject.toml")


@pytest.fixture
def pyproject_setuptools(load_files, file_types):
    files = load_files([file_types.SETUPTOOLS])
    return files[file_types.SETUPTOOLS]


@pytest.fixture
def pyproject_setuptools_file(create_files, file_types):
    folder = create_files([(file_types.SETUPTOOLS, "pyproject.toml")])
    return folder / Path("pyproject.toml")


def test_content_match(pyproject_poetry, pyproject_setuptools):
    # create a function to check both file formats
    def assert_content_match(pyproject_file):
        assert pyproject_file.name == "test-package"
        assert (
            pyproject_file.description
            == "This is a test package for demonstration purposes."
        )
        assert pyproject_file.license == "MIT"
        assert len(pyproject_file.authors) == 1

    # assert for both formats
    assert_content_match(pyproject_poetry)
    assert_content_match(pyproject_setuptools)


def test_sync(pyproject_poetry, pyproject_setuptools, somesy_input):
    def assert_sync(pyproject):
        pyproject.sync(somesy_input.project)
        assert pyproject.name == "testproject"
        assert pyproject.version == "1.0.0"

    assert_sync(pyproject_poetry)
    assert_sync(pyproject_setuptools)


def test_save(tmp_path, pyproject_poetry, pyproject_setuptools):
    def assert_save(pyproject):
        custom_path = tmp_path / Path("pyproject.toml")
        pyproject.save(custom_path)
        assert custom_path.is_file()
        custom_path.unlink()

    assert_save(pyproject_poetry)
    assert_save(pyproject_setuptools)


def test_from_to_person(person):
    # test for poetry
    assert Poetry._from_person(person) == f"{person.full_name} <{person.email}>"

    p = Poetry._to_person(Poetry._from_person(person))
    assert p.full_name == person.full_name
    assert p.email == person.email

    # test for setuptools
    assert SetupTools._from_person(person) == {
        "name": person.full_name,
        "email": person.email,
    }

    p = SetupTools._to_person(SetupTools._from_person(person))
    assert p.full_name == person.full_name
    assert p.email == person.email


@pytest.mark.parametrize(
    "writer_class, writer_file_fixture",
    [(Poetry, "pyproject_poetry_file"), (SetupTools, "pyproject_setuptools_file")],
)
def test_person_merge_pyproject(request, writer_class, writer_file_fixture, person):
    # get suitable project file
    writer_file = request.getfixturevalue(writer_file_fixture)
    pj = writer_class(writer_file)
    # update project file with known data
    pm = ProjectMetadata(
        name="My awesome project",
        description="Project description",
        license=LicenseEnum.MIT,
        version="0.1.0",
        people=[person.model_copy(update=dict(author=True, publication_author=True))],
    )
    pj.sync(pm)
    pj.save()
    # ----

    # jane becomes john -> modified person
    person1b = person.model_copy(
        update={"given_names": "John", "author": True, "publication_author": True}
    )

    # different Jane Doe with different orcid -> new person
    person2 = person.model_copy(
        update={
            "orcid": "https://orcid.org/4321-0987-3231",
            "email": "i.am.jane@doe.com",
            "author": True,
            "publication_author": True,
        }
    )
    # use different order, just for some difference
    person2.set_key_order(["given_names", "orcid", "family_names", "email"])

    # listed in "arbitrary" order in somesy metadata (new person comes first)
    pm.people = [person2, person1b]  # need to assign like that to keep _key_order
    pj.sync(pm)
    pj.save()

    # existing author info preserved, order not preserved because no orcid
    person1b_rep = writer_class._from_person(person1b)
    person2_rep = writer_class._from_person(person2)
    assert (pj.authors[0] == person1b_rep) or (pj.authors[1] == person1b_rep)
    assert (pj.authors[0] == person2_rep) or (pj.authors[1] == person2_rep)

    # new person
    person3 = Person(
        **{
            "given_names": "Janice",
            "family_names": "Doethan",
            "email": "jane93@gmail.com",
            "author": True,
            "publication_author": True,
        }
    )
    person3_rep = writer_class._from_person(person3)

    # john has a new email address
    person1c = person1b.model_copy(update={"email": "john.of.us@qualityland.com"})
    person1c_rep = writer_class._from_person(person1c)

    # jane 2 is removed from authors, but added to maintainers
    person2.author = False
    person2.publication_author = False
    person2.maintainer = True
    # reflect in project metadata
    pm.people = [person3, person2, person1c]
    # sync to CFF file
    pj.sync(pm)
    pj.save()

    assert len(pj.maintainers) == 1
    assert pj.authors[0] == person1c_rep
    assert pj.authors[1] == person3_rep



import pytest

from somesy.core.models import LicenseEnum, Person, ProjectMetadata
from somesy.package_json.writer import PackageJSON


@pytest.fixture
def package_json(load_files, file_types):
    files = load_files([file_types.PACKAGE_JSON])
    return files[file_types.PACKAGE_JSON]


@pytest.fixture
def package_json_file(create_files, file_types):
    return create_files([(file_types.PACKAGE_JSON, "package.json")]) / "package.json"


def test_content_match(package_json: PackageJSON):
    assert package_json.name == "test-package"
    assert (
        package_json.description == "This is a test package for demonstration purposes."
    )
    assert package_json.license == "MIT"
    assert len(package_json.authors) == 1


def test_sync(package_json: PackageJSON, somesy_input: dict):
    package_json.sync(somesy_input.project)
    assert package_json.name == "testproject"
    assert package_json.version == "1.0.0"


def test_save(tmp_path, package_json: PackageJSON):
    # test save with custom path
    custom_path = tmp_path / "package.json"
    package_json.save(custom_path)
    assert custom_path.is_file()


def test_from_to_person(person: Person):
    p_person = PackageJSON._from_person(person)
    assert p_person["name"] == person.full_name
    assert p_person["email"] == person.email
    assert p_person["url"] == str(person.orcid)

    p = PackageJSON._to_person(PackageJSON._from_person(person))
    assert p.full_name == person.full_name
    assert p.email == person.email
    assert p.orcid == person.orcid


def test_person_merge(package_json_file, person: Person):
    pj = PackageJSON(package_json_file)

    pm = ProjectMetadata(
        name="My awesome project",
        description="Project description",
        license=LicenseEnum.MIT,
        version="0.1.0",
        people=[person.model_copy(update=dict(author=True, publication_author=True))],
    )
    pj.sync(pm)
    pj.save()

    # jane becomes john -> modified person
    person1b = person.model_copy(
        update={"given_names": "John", "author": True, "publication_author": True}
    )

    # different Jane Doe with different orcid -> new person
    person2 = person.model_copy(
        update={
            "orcid": "https://orcid.org/4321-0987-3231",
            "email": "i.am.jane@doe.com",
            "author": True,
            "publication_author": True,
        }
    )
    # use different order, just for some difference
    person2.set_key_order(["given_names", "orcid", "family_names", "email"])

    # listed in "arbitrary" order in somesy metadata (new person comes first)
    pm.people = [person2, person1b]  # need to assign like that to keep _key_order
    pj.sync(pm)
    pj.save()

    # existing author order preserved
    assert pj.authors[0]["name"] == person1b.full_name
    assert pj.authors[0]["email"] == person1b.email

    assert pj.contributors[1]["name"] == person2.full_name
    assert pj.contributors[1]["email"] == person2.email

    # new person
    person3 = Person(
        **{
            "given_names": "Janice",
            "family_names": "Doethan",
            "email": "jane93@gmail.com",
            "author": True,
            "publication_author": True,
        }
    )
    # john has a new email address
    person1c = person1b.model_copy(update={"email": "john.of.us@qualityland.com"})
    # jane 2 is removed from authors, but added to maintainers
    person2.author = False
    person2.publication_author = False
    person2.maintainer = True
    # reflect in project metadata
    pm.people = [person3, person2, person1c]
    # sync to CFF file
    pj.sync(pm)
    pj.save()

    assert len(pj.contributors) == 3
    assert len(pj.maintainers) == 1
    assert pj.authors[0]["name"] == person1c.full_name
    assert pj.authors[0]["email"] == person1c.email

    assert pj.contributors[2]["name"] == person3.full_name
    assert pj.contributors[2]["email"] == person3.email



"""Fill command of somesy."""
import logging
from pathlib import Path
from sys import stdin

import typer
from jinja2 import Environment, FunctionLoader, select_autoescape

from .util import resolved_somesy_input, wrap_exceptions

logger = logging.getLogger("somesy")
app = typer.Typer()


@app.callback(invoke_without_command=True)
@wrap_exceptions
def fill(
    template_file: Path = typer.Option(
        None,
        "--template",
        "-t",
        exists=True,
        file_okay=True,
        dir_okay=False,
        writable=False,
        readable=True,
        resolve_path=False,
        help="Path to a Jinja2 template for somesy to fill (default: stdin).",
    ),
    input_file: Path = typer.Option(
        None,
        "--input-file",
        "-i",
        exists=True,
        file_okay=True,
        dir_okay=False,
        writable=True,
        readable=True,
        resolve_path=True,
        help="Path of somesy input file (default: try to infer).",
    ),
    output_file: Path = typer.Option(
        None,
        "--output-file",
        "-o",
        exists=False,
        file_okay=True,
        dir_okay=False,
        writable=True,
        readable=False,
        resolve_path=True,
        help="Path for target file (default: stdout).",
    ),
):
    """Fill a Jinja2 template with somesy project metadata (e.g. list authors in project docs)."""
    somesy_input = resolved_somesy_input(input_file=input_file)

    if template_file:
        logger.debug(f"Reading Jinja2 template from '{template_file}'.")
        with open(template_file, "r") as f:
            template_str = f.read()
    else:
        logger.debug("Reading Jinja2 template from stdin.")
        template_str = stdin.read()

    result = (
        Environment(
            loader=FunctionLoader(lambda _: template_str),
            autoescape=select_autoescape(),
        )
        .get_template("")
        .render(project=somesy_input.project)
    )

    if output_file:
        logger.debug(f"Writing result to '{output_file}'.")
        with open(output_file, "w") as f:
            f.write(result)
    else:
        logger.debug("Writing result to stdout.")
        print(result)



from pathlib import Path

import pytest
from tomlkit import dump, load

from somesy.cli.util import resolved_somesy_input


@pytest.fixture
def somesy_file(create_files, file_types):
    folder_path = create_files([(file_types.SOMESY, "somesy.toml")])
    file_path = Path(folder_path) / "somesy.toml"
    somesy_file = None
    with open(file_path, "r") as f:
        somesy_file = load(f)

    # delete config section and add one property
    del somesy_file["config"]
    somesy_file["config"] = {"no_sync_package_json": False}

    # save to file
    with open(file_path, "w") as f:
        dump(somesy_file, f)
    return file_path


def test_resolved_somesy_input(somesy_file):
    # set cli_args for tests
    cli_args = {"input_file": somesy_file, "debug": True, "no_sync_pyproject": True}

    somesy_input = resolved_somesy_input(**cli_args)

    # check default values
    assert somesy_input.config.no_sync_cff == False
    assert somesy_input.config.pyproject_file == Path("pyproject.toml")

    # check overwritten values
    assert somesy_input.config.no_sync_pyproject == True

    # check newly add value
    assert somesy_input.config.debug == True
